from importlib import resources as impresources
from . import demo_data
import pickle

demo_data_file = impresources.files(demo_data) / 'PPGDalia_S6_stairs.pkl'
demo_data_filename = demo_data_file.resolve()

def load_demo_data():
    """Loads demo data included in the KID-PPG package. 

    The data are taken from the PPGDalia dataset, Subject 6, Stairs activity.

    Returns:
        X (numpy.ndarray): PPG and Accelerometer signals. Size is 
          [N_samples, N_channels = 4, 256]. First channel is the PPG signal followed
          by 3 acceleration signals. 256 points correspond to 8-second windows with
          32.0Hz sampling frequency. 

        y (numpy.ndarray): Ground-truth heart rate corresponding to the windows in X.
    """
    with open(demo_data_filename, 'rb') as handle:
        data = pickle.load(handle)

    X, y = data['X'], data['y']

    return X, y


import tensorflow as tf
import keras
import numpy as np

class AdaptiveFilteringModel(keras.Model):
    """Class defining the Adaptive Linear Filtering Model.
    """
    def __init__(self, local_optimizer:tf.keras.optimizers.Optimizer, 
                 num_epochs_self_train:int = 500,
                 input_shape:tuple = (3, 256, 1), 
                 track_prediction_history:bool = False,
                 name:str = None):
        """Initializes the AdaptiveFilterModel.

        Args:
            local_optimizer (tf.keras.optimizer.Optimizer): Tensorflow optimizer to 
              train the adaptive filter.
            num_epochs_self_train (int): Number of epochs to converge the model.
              Defaults to 500.
            input_shape (tuple): Linear model input shape. Defaults to (3, 256, 1).
            track_prediction_history (bool): If True, the history of filtered signals 
              is tracked during filter training. Defaults to false.
            name (str): Name of the adaptive filter model. Defaults to None.

        Attributes:
            model (tensorflow.keras.models.Model): The adaptive linear tensorflow model.
            prediction_history (list): List of intermediate filter estimations during 
              training. Used if track_prediction_history is set to True. 
            track_prediction_history (bool): If true, prediction history is tracked.
            initial_weights (list of numpy.array): Initial random weights.
        """
        super().__init__()
        
        self.local_optimizer = local_optimizer
        self.num_epochs_self_train = num_epochs_self_train
        
        mInput = tf.keras.Input(shape = input_shape)
        
        self.conv1 = keras.layers.Conv2D(filters = 1, 
                                         kernel_size = (3, 21),
                                         padding = 'same', 
                                         activation = 'linear')
        self.conv2 = keras.layers.Conv2D(filters = 1, 
                                         kernel_size = (3, 1),
                                         padding = 'valid')
        
        m = self.conv1(mInput)
        m = self.conv2(m)
        m = m[:, 0, :, 0]
        
        self.model = keras.Model(inputs = mInput, outputs = m,
                                 name = name)
        self.initial_weights = self.model.get_weights()
        
        self.track_prediction_history = track_prediction_history
        self.prediction_history = []
        
    def reinitialize_weights(self):
        """Reset linear filter weights to initial random weights.
        """
        self.model.set_weights(self.initial_weights)
        
    def __adaptive_loss(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        """Adaptive loss function for optimizing the linear filter.

        Args:
            y_true (tensorflow.Tensor): Target (desired) signal.
            y_pred (tensorflow.Tensor): Model output signal.
        Returns:
            e (tensorflow.Tensor): Mean adaptive loss.
        """
        y_true_reshaped = y_true[:, 0, :, 0]
        y_true_fft = tf.cast(y_true_reshaped, dtype = tf.complex128)
        y_true_fft = tf.signal.fft(y_true_fft)

        
        y_pred_fft = tf.cast(y_pred, dtype = tf.complex128)
        y_pred_fft = tf.signal.fft(y_pred_fft)

        e = tf.math.abs(y_true_fft - y_pred_fft)
        e = tf.cast(e, dtype = tf.float64)
        e = tf.math.reduce_sum(tf.math.square(e), axis = -1)
                
        return tf.reduce_mean(e)
    
    def __grad(self, inputs: tf.Tensor, targets: tf.Tensor)->tuple[tf.Tensor, tf.Tensor]:
        """Calculates the required gradients for optimizing linear filter.

        Uses tensorflow's GradientTape to estimate the gradients.

        Args:
            inputs (tensorflow.Tensor): Inputs for calculating the gradient.
            targets (tensorflow.Tensor): Target tensors for calculating the gradient.
        Returns:
            loss_value (tensorflow.Tensor): Loss value based on the self.__loss() function.
            tape.gradient (tensorflow.Tensor): The calculated gradients.
        """
        with tf.GradientTape() as tape:
            loss_value = self.__loss(inputs, targets, training=True)
        return loss_value, tape.gradient(loss_value, 
                                         self.model.trainable_variables)
    
    def __loss(self, x: tf.Tensor, y: tf.Tensor, training: bool) -> tf.Tensor:
        """Helper function to calculate loss

        Args:
            x (tensorflow.Tensor): The input to the model (self.model).
            y (tensorflow.Tensor): Corresponding ground-truth values.
        
        Returns:
            loss (tensorflow.Tensor): Adaptive loss as estimated by __adaptive_loss()
              function.
        """
        y_ = self.model(x, training = training)
        
        return self.__adaptive_loss(y_true = y, y_pred = y_)        
    
    def call(self, inputs: np.ndarray, train_weights:bool = True) -> tf.Tensor:
        """Filter input PPG by using ACC signals as motion reference.

        Args:
            inputs (numpy.ndarray): Input signal composing of PPG and acceleration 
              reference signals. Input size is [N_samples, 4, 256].
            train_weights (bool): True if model weights should be adapted/trained 
              on the given input data or not. Defaults to True.
        Returns:
            x_out (tensorflow.Tensor): PPG after adaptively filtering motion artifacts 
              linearly correlated to the acceleration.
        """
        x = inputs[:, 1:, ...]
        y = inputs[:, :1, ...]
        
        self.model.trainable = train_weights

        if train_weights:
            self.reinitialize_weights()
            for epoch in range(self.num_epochs_self_train):
                loss_value, grads = self.__grad(x, y)
                self.local_optimizer.apply_gradients(zip(grads, 
                                                        self.model.trainable_variables))
                
                if self.track_prediction_history:
                    x_out = y[:, 0, :, 0] - self.model(x)
                    self.prediction_history.append(x_out)
            self.model.trainable = False

        x_out = y[:, 0, :, 0] - self.model(x)

        return x_out
    


import numpy as np


def create_temporal_pairs(X_in: np.ndarray, y_in: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Create pairs from the data X associated to labels y.
    
    Pairs of [X(n), X(n + 1)] are associated to labels y(n + 1).
    
    Args:
        X_in (numpy.ndarray): Data of shape (n, ...).
        y_in (numpy.ndarray): Labels of shape (n,).

    Returns:
        temp_X (numpy.ndarray): Pairs of input data of shape (n-1, ..., 2).
        temp_y (numpy.ndarray): Associated labels of shape (n-1,).
    """
    
    temp_X = np.concatenate([X_in[:-1, ...][..., None], 
                             X_in[1:, ...][..., None]], axis = -1)
    temp_y = y_in[1:]
    
    return temp_X, temp_y

def sample_wise_z_score_normalization(X: np.ndarray):
    """Sample-wise normalization of X.

    Args:
        X (numpy.ndarray): Data of shape (n, )
    Returns:
        X (numpy.ndarray): Sample-wise normalized output data.
        ms (numpy.ndarray): Averages used for the normalization.
        stds (numpy.ndarray): Standard deviations used for the normalization.
    """
    
    ms = np.zeros((X.shape[0], 4))
    stds = np.zeros((X.shape[0], 4))
    for i in range(X.shape[0]):
        curX = X[i, ...]
        
        for j in range(4): 
            std = np.std(curX[j, ...])
            m = np.mean(curX[j, ...])
            
            curX[j, ...] = curX[j, ...] - np.mean(curX[j, ...])
            
            if std != 0:
                curX[j, ...] = curX[j, ...] / std
            
            ms[i, j] = m
            stds[i, j] = std
        
        X[i, ...] = curX
        
    return X, ms, stds

def sample_wise_z_score_denormalization(X, ms, stds):
    """Sample-wise denormalization of X. 

    Args:
        X (numpy.ndarray): Sample-wise normalized output data.
        ms (numpy.ndarray): Averages used for the normalization.
        stds (numpy.ndarray): Standard deviations used for the normalization.
    Returns:
        X (numpy.ndarray)
    """
    
    for i in range(X.shape[0]):
        curX = X[i, ...]
        
        for j in range(X.shape[1]): 
            
            if stds[i, j] != 0:
                curX[j, ...] = curX[j, ...] * stds[i, j]
            
            curX[j, ...] = curX[j, ...] + ms[i, j]
        X[i, ...] = curX
    
    return X



import tensorflow as tf
import tensorflow_probability as tfp
import scipy
import numpy as np

from importlib import resources as impresources
from . import model_weights

tfd = tfp.distributions

weights_file = impresources.files(model_weights) / 'kid_ppg_weights.h5'
weights_filename = weights_file.resolve()

class KID_PPG:
    """KID-PPG class. Defines the KID-PPG model. 

    Attributes:
        input_shape (tuple): KID-PPG model input shape.
        model (tensorflow.keras.models.Model): Tensorflow model.
        submodel (tensorflow.keras.models.Model): Same as model but without the last
        distribution layer. 
    """
    def __init__(self, input_shape:tuple = (256, 2),
                 input_weights_file:str = None, load_weights:bool = True):
        """Initializes KID-PPG

        Args:
            input_shape (tuple): Shape of the input data given to the KID-PPG model. 
              Defaults to (256, 2).
            input_weights_file (str): String directory pointing to the location of 
              the KID-PPG model pretrained weights. Defaults to None, default location 
              with pretrained weights is used. 
            load_weights (bool): True if pretrained weights should be loaded. 
              Defaults to True.
        """
        self.input_shape = input_shape

        self.model = self.__build_model_probabilistic()

        if load_weights:

            if input_weights_file is None:
                self.model.load_weights(weights_filename)
            else:
                self.model.load_weights(input_weights_file)
        
        self.submodel = tf.keras.models.Model(inputs = self.model.inputs, 
                                              outputs = self.model.layers[-2].output)


    def predict(self, x: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        """ Estimates HR probability given PPG input.

        PPG input should be sampled at 32Hz and converted into pairs [PPG(n), PPG(n + 1)].

        Args:
            x (numpy.ndarray): PPG signal for heart rate extraction. 
            Size should be [N_samples, 256, 2].
        Returns:
            y_pred_m (numpy.ndarray): Expected HR as estimated by KID-PPG model.
              Size is [N_samples, 1].
            y_pred_std (numpy.ndarray): Estimated standard deviation of the HR distribution. 
              Size is [N_samples, 1]
        """
        y_pred = self.submodel.predict(x)
        
        y_pred_m = y_pred[:, 0]
        y_pred_std = (1 + tf.math.softplus(y_pred[:,1:2])).numpy().flatten()
        
        return y_pred_m, y_pred_std

    def predict_threshold(self, x: np.array, threshold: np.float32) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """ Estimates HR probability given PPG input. It also calculates
        the probability of error: p(error < threshold).

        PPG input should be sampled at 32Hz and converted into pairs [PPG(n), PPG(n + 1)].

        Args:
            x (numpy.ndarray): PPG signal for heart rate extraction. 
              Size should be [N_samples, 256, 2].
            threshold (numpy.float32): Threshold for estimating the probability 
              p(error > threshold).
        Returns:
            y_pred_m (numpy.ndarray): Expected HR as estimated by KID-PPG model. 
              Size is [N_samples, 1].
            y_pred_std (numpy.ndarray): Estimated standard deviation of the 
              HR distribution. Size is [N_samples, 1].
            p_error (numpy.ndarray): Estimated probability of error > threshold. 
              Size is [N_samples, 1].
        """
        y_pred_m, y_pred_std = self.predict(x)

        p_error = scipy.stats.norm(y_pred_m, y_pred_std).cdf(y_pred_m + threshold) \
                - scipy.stats.norm(y_pred_m, y_pred_std).cdf(y_pred_m - threshold)
            
        return y_pred_m, y_pred_std, p_error

    def __convolution_block(self, input_shape:tuple, n_filters:int, 
                        kernel_size:int = 5, 
                        dilation_rate:int = 2,
                        pool_size:int = 2,
                        padding:str = 'causal')->tf.keras.models.Model:
        """ Creates a convolutional block containing convolutional layers
        followed by a 1D Average Pooling and a Dropout layer.

        Args:
            input_shape (tuple): the input shape of the first convolutional layer.
            n_filters (int): Number of filters on each convolutional layer.
            kernel_size (int): Size of convolution kernel. Defautls to 5.
            dilation_rate (int): Dilation rate used in the convolution. Defaults to 2.
            pool_size (int): Pool size of the average pooling layer. Defaults to 2.
            padding (str): Padding mode used by convolutional layers. Defaults to 'str'.

        Returns:
            model (tensorflow.models.Model): Tensorflow model ofthe convolutional block.
        """
            
        mInput = tf.keras.Input(shape = input_shape)
        m = mInput
        for i in range(3):
            m = tf.keras.layers.Conv1D(filters = n_filters,
                                    kernel_size = kernel_size,
                                    dilation_rate = dilation_rate,
                                        padding = padding,
                                    activation = 'relu')(m)
        
        m = tf.keras.layers.AveragePooling1D(pool_size = pool_size)(m)
        m = tf.keras.layers.Dropout(rate = 0.5)(m, training = False)
            
        model = tf.keras.models.Model(inputs = mInput, outputs = m)
        
        return model

    def __my_dist(self, params:tf.Tensor) -> tfp.distributions.Normal:
        """Defines the Normal distribution used as the exit of the KID-PPG model.

        Args:
            params (tf.Tensor): Parameters for the Normal distribution. Shape should be 
            (..., 2), where (..., 0) refers to the loc parameters and (..., 1) to the 
            scale. 

        Returns:
            tfd.Normal() (tfp.distributions.Normal): The normal distribution.
        """
        return tfd.Normal(loc=params[:,0:1], 
                        scale = 1 + tf.math.softplus(params[:,1:2]))
        

    def __build_model_probabilistic(self, return_attention_weights:bool = False) -> tf.keras.models.Model:
        """Create probabilistic model.
    
        Model takes as inputs PPG pairs of [X(n), X(n + 1)] are associated 
        and outputs a heart rate probability estimate.
        
        Args:
            return_attention_weights (bool): Set to true if the model should return 
            the attention weights. Defaults to false.

        Returns:
            model (tensorflow.models.Model): Tensorflow model of KID-PPG.
        """
        
        modal_input_shape = (self.input_shape[0], 1)
        
        mInput = tf.keras.Input(shape = self.input_shape)
        
        mInput_t_1 = mInput[..., :1]
        mInput_t = mInput[..., 1:]
        
        conv_block1 = self.__convolution_block(modal_input_shape, n_filters = 32,
                                        pool_size = 4)
        conv_block2 = self.__convolution_block((64, 32), n_filters = 48)
        conv_block3 = self.__convolution_block((32, 48), n_filters = 64)
        
        m_ppg_t_1 = conv_block1(mInput_t_1)
        m_ppg_t_1 = conv_block2(m_ppg_t_1)
        m_ppg_t_1 = conv_block3(m_ppg_t_1)
        
        m_ppg_t = conv_block1(mInput_t)
        m_ppg_t = conv_block2(m_ppg_t)
        m_ppg_t = conv_block3(m_ppg_t)
        
        
        attention_layer = tf.keras.layers.MultiHeadAttention(num_heads = 4,
                                                            key_dim = 16,
                                                            )
        
        if return_attention_weights:
            m, attention_scores = attention_layer(query = m_ppg_t, value = m_ppg_t_1, return_attention_scores=True)
        else:
            m = attention_layer(query = m_ppg_t, value = m_ppg_t_1, return_attention_scores = False)
        
        m = m + m_ppg_t
        
        m = tf.keras.layers.LayerNormalization()(m)
        
            
        m = tf.keras.layers.Flatten()(m)
        m = tf.keras.layers.Dense(units = 256, activation = 'relu')(m)
        m = tf.keras.layers.Dropout(rate = 0.125)(m)
        m = tf.keras.layers.Dense(units = 2)(m)
        
        m = tfp.layers.DistributionLambda(self.__my_dist)(m)
        
        if return_attention_weights:
            model = tf.keras.models.Model(inputs = mInput, outputs = [m, attention_scores])
        else:
            model = tf.keras.models.Model(inputs = mInput, outputs = m)
                    
        return model 

    def summary(self):
        """ Prints model summary.
        """
        self.model.summary()